package eu.royalsloth.jmhtest;

import org.openjdk.jmh.annotations.Benchmark;
import org.openjdk.jmh.infra.Blackhole;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

/**
 * Code that was used to compare the performance of object oriented and data oriented approach to the ant
 * colony problem, mentioned in the "compiler will optimize that away" article.
 *
 * oo - object oriented
 * do - data oriented
 */
public class ooVsDoBenchmark {
    static final int ANT_COLONY_SIZE = 100;

    static class Ant {
        public String name = "unknownAnt";
        public String color = "red";
        public boolean isWarrior = false;
        public int age = 0;
    }

    static class AntColony {
        public int size = 0;
        public String[] names = new String[ANT_COLONY_SIZE];
        public String[] colors = new String[ANT_COLONY_SIZE];
        public int[] ages = new int[ANT_COLONY_SIZE];
        public boolean[] warriors = new boolean[ANT_COLONY_SIZE];
        // ^ I am aware of the fact that this array could be removed
        // by splitting the colony in two (warriors, non warriors),
        // but that is not the point of this story.
    }

    public static String createAntColor() {
        int random = ThreadLocalRandom.current().nextInt(2);
        if (random == 0) {
            return "red";
        }
        return "black";
    }

    public static int createAntAge() {
        return ThreadLocalRandom.current().nextInt(3);
    }

    public static boolean isAntWarrior() {
        return ThreadLocalRandom.current().nextInt(2) > 0;
    }

    public static List<Ant> createAntColony_oo() {
        int size = ANT_COLONY_SIZE;
        ArrayList<Ant> ants = new ArrayList<>(size);
        for (int i = 0; i < size; i++) {
            Ant ant = new Ant();
            ant.age = createAntAge();
            ant.color = createAntColor();
            ant.isWarrior = isAntWarrior();
            ants.add(ant);
        }
        return ants;
    }

    public static AntColony createAntColony_do() {
        AntColony colony = new AntColony();
        int size = colony.names.length;
        for (int i = 0; i < size; i++) {
            colony.names[i] = "unknownAnt";
            colony.ages[i] = createAntAge();
            colony.colors[i] = createAntColor();
            colony.warriors[i] = isAntWarrior();
            colony.size++;
        }
        return colony;
    }


    //////////////////////
    // Benchmark code
    //////////////////////
    static final List<Ant> antColony_oo = createAntColony_oo();
    static final AntColony antColony_do = createAntColony_do();

    @Benchmark
    public void countWarriors_oo(Blackhole blackhole) {
        long numOfWarriors = 0;
        for (Ant ant : antColony_oo) {
            if (ant.isWarrior) {
                numOfWarriors++;
            }
        }
        blackhole.consume(numOfWarriors);
    }

    @Benchmark
    public void countWariors_do(Blackhole blackhole) {
        long numOfWarriors = 0;
        for (int i = 0; i < antColony_do.size; i++) {
            boolean isWarrior = antColony_do.warriors[i];
            if (isWarrior) {
                numOfWarriors++;
            }
        }
        blackhole.consume(numOfWarriors);
    }
}
